# NOTE: This is experimental, and not used for now
import copy

from balans.base_state import _State


# def sort_list_with_indices(lst):
#     # Enumerate the list to keep track of original indices
#     indexed_list = [(val, idx) for idx, val in enumerate(lst)]

#     # Sort the indexed list based on the values
#     sorted_list = sorted(indexed_list, key=lambda x: x[0], reverse=True)

#     # Extract the sorted values and indices
#     sorted_values = [val for val, _ in sorted_list]
#     sorted_indices = [idx for _, idx in sorted_list]

#     return sorted_values, sorted_indices

def local_branching_relax(current: _State, rnd_state, delta) -> _State:
    print("*** Operator: ", "LB relax")
    print("\t Destroy current objective:", current.obj_val)
    next_state = copy.deepcopy(current)
    next_state.reset_solve_settings()

    # This is SCIP specific
    # variables = current.instance.mip.getVars()
    # binary_indexes = current.instance.binary_indexes
    # local_branching_size = int(len(binary_indexes) * delta)
    # zero_binary_vars, one_binary_vars = split_binary_vars(variables,
    #                                                       binary_indexes, current.index_to_val)
    #
    # # if current binary var is 0, flip to 1 consumes 1 unit of budget
    # # if current binary var is 1, flip to 0 consumes 1 unit of budget by (1-x)
    # zero_expr = quicksum(zero_var for zero_var in zero_binary_vars)
    # one_expr = quicksum(1 - one_var for one_var in one_binary_vars)
    # con = current.instance.model.addCons(zero_expr + one_expr <= local_branching_size)
    #
    # int_index = []
    # bin_index = []
    # count = 0
    # for var in variables:
    #     if var.vtype() == Constants.integer:
    #         current.instance.model.chgVarType(var, Constants.continuous)
    #         int_index.append(count)
    #     if var.vtype() == Constants.binary:
    #         current.instance.model.chgVarType(var, Constants.continuous)
    #         bin_index.append(count)
    #     count += 1
    #
    # current.instance.model.optimize()
    # if current.instance.model.getNSols() == 0 or current.instance.model.getStatus() == "infeasible":
    #     current.instance.model.freeTransform()
    #     count = 0
    #     for var in variables:
    #         if count in int_index:
    #             current.instance.model.chgVarType(var, Constants.integer)
    #         if count in bin_index:
    #             current.instance.model.chgVarType(var, Constants.binary)
    #         count += 1
    #     current.instance.model.delCons(con)
    #     return next_state
    #
    # lp_index_to_val, lp_obj_val = get_index_to_val_and_objective(current.instance.model)
    #
    # # Get back the original model
    # current.instance.model.freeTransform()
    # count = 0
    # for var in variables:
    #     if count in int_index:
    #         current.instance.model.chgVarType(var, Constants.integer)
    #     if count in bin_index:
    #         current.instance.model.chgVarType(var, Constants.binary)
    #     count += 1
    # current.instance.model.delCons(con)
    #
    # #  If a discrete variable x_rand1 = x_inc, do not change it.
    # indexes_with_same_value = [i for i in binary_indexes if
    #                            math.isclose(lp_index_to_val[i], current.index_to_val[i])]
    #
    # # Else potentially change it
    # indexes_with_diff_value = [i for i in binary_indexes if i not in indexes_with_same_value]
    # size = int(delta * len(indexes_with_diff_value))
    # next_state.destroy_set = set(rnd_state.choice(indexes_with_diff_value, size))
    #
    # #     zero_set = set([i for i in binary_indexes if i in indexes_with_same_value])
    # #     if len(non_zero_set) >= local_branching_size:
    # #         diff = [abs(lp_index_to_val[i] - current.index_to_val[i]) if i in binary_indexes else 0
    # #                 for i in range(len(lp_index_to_val))]
    # #         diff_value, diff_index = sort_list_with_indices(diff)
    # #         lb_relax_set = set(diff_index[:local_branching_size])
    # #     else:
    # #         random_set = set(rnd_state.choice(list(zero_set),
    # #         local_branching_size - len(non_zero_set), replace=False))
    # #         lb_relax_set = non_zero_set | random_set

    return next_state


def local_branching_relax_10(current: _State, rnd_state) -> _State:
    return local_branching_relax(current, rnd_state, delta=0.1)


def local_branching_relax_25(current: _State, rnd_state) -> _State:
    return local_branching_relax(current, rnd_state, delta=0.25)


import copy

from balans.base_state import _State


def dins(current: _State, rnd_state) -> _State:
    #  Take an LP relaxed solution of the original MIP.
    #  By considering only discrete variables, forms a Set J where |x_lp -x_inc| >= 0.5
    #  IF delta is given, do local branching for binary variables, change at most delta of them.
    #  If a variable is inside the Set J, it is part of the destroy set.
    #  Send the dins set (Set J) and local branching size to base_instance.

    print("*** Operator: ", "DINS")
    print("\t Destroy current objective:", current.obj_val)
    next_state = copy.deepcopy(current)
    next_state.reset_solve_settings()

    # Static features from the instance
    integer_indexes = current.instance.integer_indexes
    lp_index_to_val = current.instance.lp_index_to_val

    # DINS for integer variables: form a set_j where |x_lp - x_inc| >= 0.5
    next_state.dins_set = set([i for i in integer_indexes
                               if abs(lp_index_to_val[i] - current.index_to_val[i]) >= 0.5])

    return next_state



from copy import deepcopy
from typing import Any, Dict

from balans.base_instance import _Instance


class _State:
    """
    State of an instance with its solution and objective
    First, operators sets the solve settings of the state object
    Then, state calls instance.solve() with settings dictated by the operators
    """

    def __init__(self,
                 instance: _Instance,
                 index_to_val: Dict[Any, float],
                 obj_val: float,
                 destroy_set=None,
                 dins_set=None,
                 rens_float_set=None,
                 has_random_obj=False,
                 previous_index_to_val=None,
                 local_branching_size=0,
                 proximity_delta=None):

        # State holds an instance and its solution and objective
        # Instance holds features and the solve() logic dictated by operators
        self.instance = instance
        self.index_to_val = index_to_val  # index defined by SCIP var.getIndex()
        self.obj_val = obj_val

        # State receives the solve settings from the operators
        # and passes to instance.solve() and updates its solution and objective
        self.destroy_set = destroy_set
        self.dins_set = dins_set
        self.rens_float_set = rens_float_set
        self.has_random_obj = has_random_obj
        self.previous_index_to_val = previous_index_to_val
        self.local_branching_size = local_branching_size
        self.proximity_delta = proximity_delta

    def solution(self):
        return self.index_to_val

    def objective(self):
        return self.obj_val

    def reset_solve_settings(self):
        self.destroy_set = None
        self.dins_set = None
        self.rens_float_set = None
        self.has_random_obj = False
        self.local_branching_size = 0
        self.proximity_delta = False

    def solve_and_update(self):
        # Solve the current state with the destroyed variables and update solution and objective
        self.index_to_val, self.obj_val = self.instance.solve(index_to_val=self.index_to_val,
                                                              obj_val=self.obj_val,
                                                              destroy_set=self.destroy_set,
                                                              dins_set=self.dins_set,
                                                              rens_float_set=self.rens_float_set,
                                                              has_random_obj=self.has_random_obj,
                                                              local_branching_size=self.local_branching_size,
                                                              proximity_delta=self.proximity_delta)

    def __deepcopy__(self, memo):
        # No need to copy the instance when create a deepcopy of state object, share the instance between the copy
        shared_attribute_names = ["instance"]
        assert isinstance(shared_attribute_names, (list, tuple))
        shared_attributes = {k: getattr(self, k) for k in shared_attribute_names}

        deepcopy_method = None

        if hasattr(self, '__deepcopy__'):
            # Do hack to prevent infinite recursion in call to deepcopy
            deepcopy_method = self.__deepcopy__
            self.__deepcopy__ = None

        for attr in shared_attribute_names:
            del self.__dict__[attr]

        clone = deepcopy(self)

        for attr, val in shared_attributes.items():
            setattr(self, attr, val)
            setattr(clone, attr, val)

        if hasattr(self, '__deepcopy__'):
            # Undo hack
            self.__deepcopy__ = deepcopy_method
            del clone.__deepcopy__

        return clone



import copy
import math

from balans.base_state import _State


def _rins(current: _State, rnd_state, delta) -> _State:
    #  Take the LP relaxed solution of the original MIP and the current incumbent.
    #  If a DISCRETE variable x_inc = x_lp, then do not change it.
    #  Otherwise, put it to the destroy set.
    #  Send the destroy set to base_instance.

    print("*** Operator: ", "RINS")
    print("\t Destroy current objective:", current.obj_val)
    next_state = copy.deepcopy(current)
    next_state.reset_solve_settings()

    # Static features from the instance
    discrete_indexes = current.instance.discrete_indexes
    lp_index_to_val = current.instance.lp_index_to_val

    #  If a variable x_cur = x_lp, do not change it.
    indexes_with_diff_value = [i for i in discrete_indexes if not
    math.isclose(lp_index_to_val[i], current.index_to_val[i])]

    # Else potentially change it
    size = int(delta * len(indexes_with_diff_value))
    next_state.destroy_set = set(rnd_state.choice(indexes_with_diff_value, size))

    return next_state


def rins_05(current: _State, rnd_state) -> _State:
    return _rins(current, rnd_state, delta=0.05)


def rins_10(current: _State, rnd_state) -> _State:
    return _rins(current, rnd_state, delta=0.10)


def rins_15(current: _State, rnd_state) -> _State:
    return _rins(current, rnd_state, delta=0.15)


def rins_20(current: _State, rnd_state) -> _State:
    return _rins(current, rnd_state, delta=0.20)


def rins_25(current: _State, rnd_state) -> _State:
    return _rins(current, rnd_state, delta=0.25)


def rins_30(current: _State, rnd_state) -> _State:
    return _rins(current, rnd_state, delta=0.30)


def rins_35(current: _State, rnd_state) -> _State:
    return _rins(current, rnd_state, delta=0.35)


def rins_40(current: _State, rnd_state) -> _State:
    return _rins(current, rnd_state, delta=0.40)


def rins_45(current: _State, rnd_state) -> _State:
    return _rins(current, rnd_state, delta=0.45)


def rins_50(current: _State, rnd_state) -> _State:
    return _rins(current, rnd_state, delta=0.50)


def rins_55(current: _State, rnd_state) -> _State:
    return _rins(current, rnd_state, delta=0.55)


def rins_60(current: _State, rnd_state) -> _State:
    return _rins(current, rnd_state, delta=0.60)


def rins_65(current: _State, rnd_state) -> _State:
    return _rins(current, rnd_state, delta=0.65)


def rins_70(current: _State, rnd_state) -> _State:
    return _rins(current, rnd_state, delta=0.70)


def rins_75(current: _State, rnd_state) -> _State:
    return _rins(current, rnd_state, delta=0.75)


def rins_80(current: _State, rnd_state) -> _State:
    return _rins(current, rnd_state, delta=0.80)


def rins_85(current: _State, rnd_state) -> _State:
    return _rins(current, rnd_state, delta=0.85)


def rins_90(current: _State, rnd_state) -> _State:
    return _rins(current, rnd_state, delta=0.90)


def rins_95(current: _State, rnd_state) -> _State:
    return _rins(current, rnd_state, delta=0.95)



import copy

from balans.base_state import _State


def _rens(current: _State, rnd_state, delta) -> _State:
    #  Take the LP relaxed solution of the original MIP.
    #  For discrete variables, choose the non-integral ones
    #  and store them in the rens_float_set to be bounded list.
    #  Send the destroy set (None for the rens case) and
    #  rens_float_set to base_instance.

    print("*** Operator: ", "RENS")
    print("\t Destroy current objective:", current.obj_val)
    next_state = copy.deepcopy(current)
    next_state.reset_solve_settings()

    # Static features from the instance
    floating_discrete_indexes = current.instance.lp_floating_discrete_indexes

    # Randomization STEP
    size = int(delta * len(floating_discrete_indexes))
    next_state.rens_float_set = set(rnd_state.choice(floating_discrete_indexes, size))

    return next_state


def rens_05(current: _State, rnd_state) -> _State:
    return _rens(current, rnd_state, delta=0.05)


def rens_10(current: _State, rnd_state) -> _State:
    return _rens(current, rnd_state, delta=0.10)


def rens_15(current: _State, rnd_state) -> _State:
    return _rens(current, rnd_state, delta=0.15)


def rens_20(current: _State, rnd_state) -> _State:
    return _rens(current, rnd_state, delta=0.20)


def rens_25(current: _State, rnd_state) -> _State:
    return _rens(current, rnd_state, delta=0.25)


def rens_30(current: _State, rnd_state) -> _State:
    return _rens(current, rnd_state, delta=0.30)


def rens_35(current: _State, rnd_state) -> _State:
    return _rens(current, rnd_state, delta=0.35)


def rens_40(current: _State, rnd_state) -> _State:
    return _rens(current, rnd_state, delta=0.40)


def rens_45(current: _State, rnd_state) -> _State:
    return _rens(current, rnd_state, delta=0.45)


def rens_50(current: _State, rnd_state) -> _State:
    return _rens(current, rnd_state, delta=0.50)


def rens_55(current: _State, rnd_state) -> _State:
    return _rens(current, rnd_state, delta=0.55)


def rens_60(current: _State, rnd_state) -> _State:
    return _rens(current, rnd_state, delta=0.60)


def rens_65(current: _State, rnd_state) -> _State:
    return _rens(current, rnd_state, delta=0.65)


def rens_70(current: _State, rnd_state) -> _State:
    return _rens(current, rnd_state, delta=0.70)


def rens_75(current: _State, rnd_state) -> _State:
    return _rens(current, rnd_state, delta=0.75)


def rens_80(current: _State, rnd_state) -> _State:
    return _rens(current, rnd_state, delta=0.80)


def rens_85(current: _State, rnd_state) -> _State:
    return _rens(current, rnd_state, delta=0.85)


def rens_90(current: _State, rnd_state) -> _State:
    return _rens(current, rnd_state, delta=0.90)


def rens_95(current: _State, rnd_state) -> _State:
    return _rens(current, rnd_state, delta=0.95)



import os
from typing import NamedTuple

import mabwiser.utils


class Constants(NamedTuple):
    """
    Constant values used by the modules.
    """

    # Default seed
    default_seed = 123456

    # Default MIP Solver
    scip_solver = "scip"
    gurobi_solver = "gurobi"
    highs_solver = "highs"
    default_solver = scip_solver

    # Optimization sense
    minimize = "minimize"
    maximize = "maximize"

    # Scip variable types
    binary = "BINARY"
    integer = "INTEGER"
    continuous = "CONTINUOUS"

    # Column names for features df
    var_type = "var_type"
    var_lb = "var_lb"
    var_ub = "var_ub"

    # Time limit for the initial solution to get feasible solution as a starting point for ALNS
    timelimit_first_solution = 20

    # Time limit for finding a random feasible solution
    timelimit_random_feasible = 20

    # time limit for one iteration is ALNS, local branching has longer time because hard problem created
    timelimit_alns_iteration = 60

    # time limit for one local branching iteration.
    # paper says Each LNS iteration is limited to 1 minute, except for Local Branching with 2.5 minutes.
    timelimit_local_branching_iteration = 150

    # for Big-M constraint, currently used in Proximity
    M = 1000

    # Data folder constants
    _TEST_DIR_NAME = "tests"
    _DATA_DIR_NAME = "data"
    _TEST_DATA_DIR_NAME = _TEST_DIR_NAME + os.sep + _DATA_DIR_NAME

    # Data paths
    _FILE_DIR = os.path.dirname(os.path.abspath(__file__))
    DATA_TEST = _FILE_DIR + os.sep + ".." + os.sep + _TEST_DATA_DIR_NAME


def create_rng(seed):
    return mabwiser.utils.create_rng(seed)


def check_false(expression: bool, exception: Exception):
    return mabwiser.utils.check_false(expression, exception)


def check_true(expression: bool, exception: Exception):
    return mabwiser.utils.check_true(expression, exception)



import copy

from balans.base_state import _State


def _local_branching(current: _State, rnd_state, delta) -> _State:
    #  For binary variables we have a hard constraint,
    #  here we say change at most half of them (delta=0.5).
    #  Other possible delta values are 0.25 and 0.75.
    #  Send the destroy set to base_instance.

    print("*** Operator: ", "LOCAL BRANCHING")
    print("\t Destroy current objective:", current.obj_val)
    next_state = copy.deepcopy(current)
    next_state.reset_solve_settings()

    # Static features from the instance
    binary_indexes = current.instance.binary_indexes

    # <= k in local branching
    next_state.local_branching_size = rnd_state.randint(int(0.05 * len(binary_indexes)),
                                                        int(delta * len(binary_indexes)) + 1)

    return next_state


def local_branching_05(current: _State, rnd_state) -> _State:
    return _local_branching(current, rnd_state, delta=0.05)


def local_branching_10(current: _State, rnd_state) -> _State:
    return _local_branching(current, rnd_state, delta=0.10)


def local_branching_15(current: _State, rnd_state) -> _State:
    return _local_branching(current, rnd_state, delta=0.15)


def local_branching_20(current: _State, rnd_state) -> _State:
    return _local_branching(current, rnd_state, delta=0.20)


def local_branching_25(current: _State, rnd_state) -> _State:
    return _local_branching(current, rnd_state, delta=0.25)


def local_branching_30(current: _State, rnd_state) -> _State:
    return _local_branching(current, rnd_state, delta=0.30)


def local_branching_35(current: _State, rnd_state) -> _State:
    return _local_branching(current, rnd_state, delta=0.35)


def local_branching_40(current: _State, rnd_state) -> _State:
    return _local_branching(current, rnd_state, delta=0.40)


def local_branching_45(current: _State, rnd_state) -> _State:
    return _local_branching(current, rnd_state, delta=0.45)


def local_branching_50(current: _State, rnd_state) -> _State:
    return _local_branching(current, rnd_state, delta=0.50)


def local_branching_55(current: _State, rnd_state) -> _State:
    return _local_branching(current, rnd_state, delta=0.55)


def local_branching_60(current: _State, rnd_state) -> _State:
    return _local_branching(current, rnd_state, delta=0.60)


def local_branching_65(current: _State, rnd_state) -> _State:
    return _local_branching(current, rnd_state, delta=0.65)


def local_branching_70(current: _State, rnd_state) -> _State:
    return _local_branching(current, rnd_state, delta=0.70)


def local_branching_75(current: _State, rnd_state) -> _State:
    return _local_branching(current, rnd_state, delta=0.75)


def local_branching_80(current: _State, rnd_state) -> _State:
    return _local_branching(current, rnd_state, delta=0.80)


def local_branching_85(current: _State, rnd_state) -> _State:
    return _local_branching(current, rnd_state, delta=0.85)


def local_branching_90(current: _State, rnd_state) -> _State:
    return _local_branching(current, rnd_state, delta=0.90)


def local_branching_95(current: _State, rnd_state) -> _State:
    return _local_branching(current, rnd_state, delta=0.95)



import unittest

from balans.utils import Constants


class BaseTest(unittest.TestCase):

    mip_solver = Constants.default_solver

    def assertIsBetter(self, before, after, sense):
        if sense == Constants.minimize:
            self.assertLess(after, before)
        else:
            self.assertGreater(after, before)

    def is_not_worse(self, before, after, sense):
        if sense == Constants.minimize:
            self.assertLessEqual(after, before)
        else:
            self.assertGreaterEqual(after, before)
